#pragma once

#include<dune/common/parametertreeparser.hh>

#include<dune/pdelab/finiteelementmap/p0fem.hh>
#include <dune/pdelab/adaptivity/adaptivity.hh>
#include<dune/pdelab/gridoperator/gridoperator.hh>
#include<dune/pdelab/backend/istl.hh>

#include <dune/phasefield/base_parameters.hh>
#include <dune/phasefield/debug/vector_debug.hh>

//_________________________________________________________________________________________________________________________

template<typename Grid>
class chns_base
{
public:
  typedef typename Grid::LeafGridView GV;
  const int dim = GV::dimension;
  typedef double RF;


  chns_base(Grid& grid, Dune::ParameterTree& ptree)
  {
  }

};

class TimestepManager
{
public:
  double T;

  Params_Time& time;

  double dtMin;
  double dtMax;
  double dtIncreaseFactor;
  double dtDecreaseFactor;

  double lastsave;
  double saveintervall;
  double saveintervallstart;

  int verb;

  TimestepManager(const Dune::ParameterTree & param, Params_Time& time_, int verbosity) :
  time(time_)
  {
    verb = verbosity;
    T=param.get("tMax",(double)1.0);
    time.dt=param.get("dt",(double)0.1);

    dtMin = param.get("dtMin",(double)time.dt);
    dtMax = param.get("dtMax",(double)time.dt);
    dtIncreaseFactor = param.get("dtIncreaseFactor",(double)1.2);
    dtDecreaseFactor = param.get("dtDecreaseFactor",(double)0.5);

    saveintervall = param.get("saveintervall",(double)0.0);
    saveintervallstart = param.get("saveintervallstart",(double)0.0);

    time.t=param.get("tIni",(double)0.0);
    lastsave = time.t - saveintervall;
  }

  // RETURN VALUE: -1 -> ABORT, 0-> REDO TIMESTEP, 1-> OK
  template<typename NS_NEWTON, typename CH_NEWTON, typename VTKWRITER>
  int doTimestep(NS_NEWTON& ns_newton, CH_NEWTON& ch_newton, VTKWRITER& vtkwriter)
  {
    try
    {
      if (verb > 1)   std::cout << "= Begin newtonapply NS:" << std::endl;
      ns_newton.apply();

      if (verb > 1)   std::cout << "= Begin newtonapply CH:" << std::endl;
      ch_newton.apply();

    }
    catch(...)
    {
      return rejectTimestep();
    }

    applyTimestep(vtkwriter);

    return 1;
  }

  // RETURN VALUE: -1 -> ABORT, 0-> REDO TIMESTEP, 1-> OK
  template<typename NS_NEWTON, typename VTKWRITER>
  int doOnlyNSTimestep(NS_NEWTON& ns_newton, VTKWRITER& vtkwriter)
  {
    try
    {
      if (verb > 1)   std::cout << "= Begin (only) newtonapply NS:" << std::endl;
      ns_newton.apply();
    }
    catch(...)
    {
      return rejectTimestep();
    }
    applyTimestep(vtkwriter);
    return 1;
  }

  // RETURN VALUE: -1 -> ABORT, 0-> REDO TIMESTEP, 1-> OK
  template<typename CH_NEWTON, typename VTKWRITER>
  int doOnlyCHTimestep(CH_NEWTON& ch_newton, VTKWRITER& vtkwriter)
  {
    try
    {
      if (verb > 1)   std::cout << "= Begin (only) newtonapply CH:" << std::endl;
      ch_newton.apply();
    }
    catch(...)
    {
      return rejectTimestep();
    }
    applyTimestep(vtkwriter);
    return 1;
  }

  template<typename NS_NEWTON, typename CH_NEWTON, typename VTKWRITER, typename NS_V, typename CH_V>
  int doTimestepToConvergence(NS_NEWTON& ns_newton, CH_NEWTON& ch_newton, VTKWRITER& vtkwriter, NS_V& nsV, CH_V& chV)
  {
    NS_V nsVLast = nsV;
    CH_V chVLast = chV;
    int iter = 0;
    try
    {
      do
      {
        if (verb > 1)   std::cout << "= NS/CH Iteration " << iter << std::endl;
        nsVLast = nsV;
        chVLast = chV;

        if (verb > 1)   std::cout << "= Begin newtonapply NS:" << std::endl;
        ns_newton.apply();

        if (verb > 1)   std::cout << "= Begin newtonapply CH:" << std::endl;
        ch_newton.apply();

        if (verb > 1) std::cout << "Result: " << ns_newton.result().iterations << " and " << ch_newton.result().iterations << std::endl;
        iter++;
      }
      while(iter < 10 && (ns_newton.result().iterations > 0 || ch_newton.result().iterations > 0));

    }
    catch(...)
    {
      return rejectTimestep();
    }

    applyTimestep(vtkwriter);

    return 1;
  }


  template<typename VTKWRITER>
  void applyTimestep(VTKWRITER& vtkwriter)
  {
    applyTimestep();
    if( time.t < saveintervallstart || time.t>lastsave+saveintervall )
    {
        lastsave = time.t;
        vtkwriter.write(time.t,Dune::VTK::ascii);
    }
  }

  void applyTimestep()
  {
    time.t += time.dt;
    if(time.dt < dtMax)
    {
      time.dt = std::min(time.dt*dtIncreaseFactor, dtMax);
      if (verb > 0)   std::cout << "New Timestep "<< time.dt << std::endl;
    }
  }

  int rejectTimestep()
  {
    if(time.dt <= dtMin)
    {
      std::cout << "===============================" << std::endl;
      if(dtMin==dtMax)
        std::cout << "Problem with Newton Solver, aborting :(" << std::endl;
      else
        std::cout << "Timestep too small, aborting :(" << std::endl;
      std::cout << "===============================" << std::endl;
      return -1;
    }
    time.dt = std::max(time.dt*dtDecreaseFactor,dtMin);
    if (verb > 0) std::cout << "Reducing Timestep to "<< time.dt << std::endl;
    return 0;
  }


  bool isFinished()
  {
    return (time.t>=T);
  }
};

//_________________________________________________________________________________________________________________________

template<typename GFS, typename NS_BDRY_FN, typename CH_BDRY_FN, typename NS_V, typename CH_V>
void interpolateToBdry(const GFS& gfs, const NS_BDRY_FN& nsBdryFn, const CH_BDRY_FN& chBdryFn, NS_V& nsV, CH_V& chV)
{
  NS_V nsInterp(gfs.ns);
  CH_V chInterp(gfs.ch);
  Dune::PDELab::interpolate(nsBdryFn, gfs.ns, nsInterp);
  Dune::PDELab::interpolate(chBdryFn, gfs.ch, chInterp);
  Dune::PDELab::copy_constrained_dofs(gfs.nsCC, nsInterp, nsV);
  Dune::PDELab::copy_constrained_dofs(gfs.chCC, chInterp, chV);
}

//_________________________________________________________________________________________________________________________

template<class Grid, class GV, class GFS1, class GFS2, class X1, class X2>
void my_adapt_grid (Grid& grid, GV& gv, GFS1& gfs1, X1& x1, GFS2& gfs2, X2& x2, int int_order)
{
  typedef Dune::PDELab::L2Projection<GFS1,X1> Projection1;
  Projection1 projection1(gfs1,int_order);
  typedef Dune::PDELab::L2Projection<GFS2,X2> Projection2;
  Projection2 projection2(gfs2,int_order);

  Dune::PDELab::GridAdaptor<Grid,GFS1,X1,Projection1> grid_adaptor1(gfs1);
  Dune::PDELab::GridAdaptor<Grid,GFS2,X2,Projection2> grid_adaptor2(gfs2);

  // prepare the grid for refinement
  grid.preAdapt();

  // save solution
  typename Dune::PDELab::GridAdaptor<Grid,GFS1,X1,Projection1>::MapType transferMap1;
  grid_adaptor1.backupData(grid,gfs1,projection1,x1,transferMap1);
  typename Dune::PDELab::GridAdaptor<Grid,GFS2,X2,Projection2>::MapType transferMap2;
  grid_adaptor2.backupData(grid,gfs2,projection2,x2,transferMap2);

  // adapt the grid

  std::cout << "Grid vor adapt";
  grid.adapt();
  std::cout << "Grid nach adapt";
  gv.update();
  std::cout << "nach GV adapt";

  // update the function spaces
  gfs1.update(true);
  gfs2.update(true);

  // interpolate solution
  x1 = X1(gfs1,0.0);
  grid_adaptor1.replayData(grid,gfs1,projection1,x1,transferMap1);
  x2 = X2(gfs2,0.0);
  grid_adaptor2.replayData(grid,gfs2,projection2,x2,transferMap2);

  // clean up
  grid.postAdapt();
}


template<class Grid, class GV, class GFS1, class GFS2, class X1, class X2>
void my_adapt_grid (Grid& grid, GV& gv, GFS1& gfs1, X1& x1, X1& x1b, GFS2& gfs2, X2& x2, X2& x2b, int int_order)
{
  typedef Dune::PDELab::L2Projection<GFS1,X1> Projection1;
  Projection1 projection1(gfs1,int_order);
  typedef Dune::PDELab::L2Projection<GFS2,X2> Projection2;
  Projection2 projection2(gfs2,int_order);

  Dune::PDELab::GridAdaptor<Grid,GFS1,X1,Projection1> grid_adaptor1(gfs1);
  Dune::PDELab::GridAdaptor<Grid,GFS1,X1,Projection1> grid_adaptor1b(gfs1);
  Dune::PDELab::GridAdaptor<Grid,GFS2,X2,Projection2> grid_adaptor2(gfs2);
  Dune::PDELab::GridAdaptor<Grid,GFS2,X2,Projection2> grid_adaptor2b(gfs2);

  // prepare the grid for refinement
  grid.preAdapt();

  // save solution
  typename Dune::PDELab::GridAdaptor<Grid,GFS1,X1,Projection1>::MapType transferMap1;
  grid_adaptor1.backupData(grid,gfs1,projection1,x1,transferMap1);
  grid_adaptor1b.backupData(grid,gfs1,projection1,x1b,transferMap1);
  typename Dune::PDELab::GridAdaptor<Grid,GFS2,X2,Projection2>::MapType transferMap2;
  grid_adaptor2.backupData(grid,gfs2,projection2,x2,transferMap2);
  grid_adaptor2b.backupData(grid,gfs2,projection2,x2b,transferMap2);

  // adapt the grid

  std::cout << "Grid vor adapt";
  grid.adapt();
  std::cout << "Grid nach adapt";
  gv.update();
  std::cout << "nach GV adapt";

  // update the function spaces
  gfs1.update(true);
  gfs2.update(true);

  // interpolate solution
  x1 = X1(gfs1,0.0);
  grid_adaptor1.replayData(grid,gfs1,projection1,x1,transferMap1);
  x1b = X1(gfs1,0.0);
  grid_adaptor1b.replayData(grid,gfs1,projection1,x1b,transferMap1);
  x2 = X2(gfs2,0.0);
  grid_adaptor2.replayData(grid,gfs2,projection2,x2,transferMap2);
  x2b = X2(gfs2,0.0);
  grid_adaptor2b.replayData(grid,gfs2,projection2,x2b,transferMap2);
  // clean up
  grid.postAdapt();
}




//_________________________________________________________________________________________________________________________
template<typename GV, typename RF, typename EST_LOP, typename EST_GFS>
class RefinementIndicator
{
public:
  //Estimator GFS
  typedef Dune::PDELab::P0LocalFiniteElementMap<RF,RF,GV::dimension> P0FEM;
  P0FEM p0fem;
  typedef Dune::PDELab::NoConstraints NCON;
  typedef Dune::PDELab::GridFunctionSpace<GV,P0FEM,NCON,Dune::PDELab::ISTL::VectorBackend<> > P0GFS;
  P0GFS p0Gfs;
  //Estimator vector
  using Z0 = Dune::PDELab::Backend::Vector<P0GFS,RF>;
  Z0 z0;

  typedef Dune::PDELab::DiscreteGridFunction<P0GFS,Z0> Z0DGF;
  Z0DGF z0dgf;

  //Estimator Operator
  EST_LOP estlop;
  typedef Dune::PDELab::EmptyTransformation NCC;
  typedef Dune::PDELab::ISTL::BCRSMatrixBackend<> MBE;
  typedef Dune::PDELab::GridOperator<EST_GFS,P0GFS,EST_LOP,MBE,RF,RF,RF,NCC,NCC> ESTGO;
  ESTGO estgo;

  double eta_refine;
  double eta_coarsen;
  int Rmaxlevel;
  int Rminlevel;

  RefinementIndicator(const GV& gv, const EST_GFS chGfs, const Dune::ParameterTree & param) :
    p0fem(Dune::GeometryTypes::simplex(GV::dimension)),
    p0Gfs(gv,p0fem),
    z0(p0Gfs,0.0),
    z0dgf(p0Gfs,z0),
    estlop(),
    estgo(chGfs,p0Gfs,estlop,MBE(1))
  {
    eta_refine = param.get("etaRefine",(double)0.15);
    eta_coarsen = param.get("etaCoarsen",(double)0.05);
    Rmaxlevel = param.get("maxRefineLevel",(double)2);
    Rminlevel = param.get("minRefineLevel",(double)0);
  }

  template<typename CH_V>
  const Z0& calculateIndicator(const GV& gv, const CH_V& chV, bool writeOutput)
  {
    estgo.residual(chV,z0);
    if(writeOutput)
    {
      auto vtkwriter = std::make_shared<Dune::SubsamplingVTKWriter<GV> >(gv,0);
      vtkwriter->addVertexData(std::make_shared<Dune::PDELab::VTKGridFunctionAdapter<Z0DGF> >(z0dgf,"indicator"));
      vtkwriter->write("RefinementIndicator",Dune::VTK::ascii);
    }
    return z0;
  }

  template<typename Grid>
  void markGrid(Grid &grid, int verbosity=0)
  {
    if(verbosity>0) std::cout<<"mark elements: ";
    Dune::PDELab::mark_grid(grid,z0,eta_refine,eta_coarsen, Rminlevel,Rmaxlevel,verbosity,-1);
  }

  template<typename Grid, typename CH_V, typename NS_V, typename CH_GFS, typename NS_GFS>
  void refineGrid(Grid &grid, CH_V &chV, CH_V &chVOld, NS_V &nsV, NS_V &nsVOld, CH_GFS chgfs, NS_GFS nsgfs, int verbosity = 0)
  {
    if(verbosity > 1)
    {
      printVector(chV,verbosity-2,"chV");
      printVector(nsV,verbosity-2,"nsV");
    }

    // do refinement
    if(verbosity > 0) std::cout << "adapt grid and solution...   ";

    auto chTransfer = transferSolutions(chgfs,4,chV,chVOld);
    auto nsTransfer = transferSolutions(nsgfs,4,nsV,nsVOld);
    Dune::PDELab::adaptGrid(grid, chTransfer,nsTransfer);

    if(verbosity > 0) std::cout << "   ...adapt ended" << std::endl;

    if(verbosity > 1)
    {

      printVector(chV,verbosity-2,"chV");
      printVector(nsV,verbosity-2,"nsV");
    }
  }

  template<typename Grid, typename CH_V, typename NS_V, typename CH_GFS, typename NS_GFS>
  void refineGrid(Grid &grid, CH_V &chV, NS_V &nsV, CH_GFS chgfs, NS_GFS nsgfs, int verbosity = 0)
  {
    if(verbosity > 1)
    {
      printVector(chV,verbosity-2,"chV");
      printVector(nsV,verbosity-2,"nsV");
    }

    // do refinement
    if(verbosity > 0) std::cout << "adapt grid and solution...   ";
    auto chTransfer = transferSolutions(chgfs,4,chV);
    auto nsTransfer = transferSolutions(nsgfs,4,nsV);
    Dune::PDELab::adaptGrid(grid, chTransfer,nsTransfer);

    if(verbosity > 0) std::cout << "   ...adapt ended" << std::endl;

    if(verbosity > 1)
    {

      printVector(chV,verbosity-2,"chV");
      printVector(nsV,verbosity-2,"nsV");
    }
  }

  template<typename Grid, typename CH_V, typename NS_V, typename CH_GFS, typename NS_GFS>
  void myRefineGrid(Grid &grid, GV& gv, CH_V &chV, NS_V &nsV, CH_GFS chgfs, NS_GFS nsgfs, int verbosity = 0)
  {
    if(verbosity > 1)
    {
      printVector(chV,verbosity-2,"chV");
      printVector(nsV,verbosity-2,"nsV");
    }

    // do refinement
    if(verbosity > 0) std::cout << "adapt grid and solution...   ";
    //auto chTransfer = transferSolutions(chgfs,4,chV);
    //auto nsTransfer = transferSolutions(nsgfs,4,nsV);
    //Dune::PDELab::adaptGrid(grid, chTransfer,nsTransfer);
    my_adapt_grid(grid, gv, chgfs, chV, nsgfs, nsV,4);

    if(verbosity > 0) std::cout << "   ...adapt ended" << std::endl;

    if(verbosity > 1)
    {

      printVector(chV,verbosity-2,"chV");
      printVector(nsV,verbosity-2,"nsV");
    }
  }

  template<typename Grid, typename CH_V, typename NS_V, typename CH_GFS, typename NS_GFS>
  void myRefineGrid(Grid &grid, GV& gv, CH_V &chV, CH_V &chVOld, NS_V &nsV, NS_V &nsVOld, CH_GFS chgfs, NS_GFS nsgfs, int verbosity = 0)
  {
    if(verbosity > 1)
    {
      printVector(chV,verbosity-2,"chV");
      printVector(nsV,verbosity-2,"nsV");
    }

    // do refinement
    if(verbosity > 0) std::cout << "adapt grid and solution...   ";
    //auto chTransfer = transferSolutions(chgfs,4,chV);
    //auto nsTransfer = transferSolutions(nsgfs,4,nsV);
    //Dune::PDELab::adaptGrid(grid, chTransfer,nsTransfer);
    my_adapt_grid(grid, gv, chgfs, chV, chVOld, nsgfs, nsV, nsVOld, 4);

    if(verbosity > 0) std::cout << "   ...adapt ended" << std::endl;

    if(verbosity > 1)
    {

      printVector(chV,verbosity-2,"chV");
      printVector(nsV,verbosity-2,"nsV");
    }
  }
};
