#pragma once

#include<dune/pdelab/finiteelement/localbasiscache.hh>
#include<dune/pdelab/localoperator/defaultimp.hh>
#include<dune/pdelab/localoperator/variablefactories.hh>
#include<dune/pdelab/gridfunctionspace/gridfunctionspace.hh>

template<typename GFS, typename Container, typename Entity, typename RangeFieldType>
class GFS_LocalViewer
{
public:
  typedef Dune::PDELab::LocalFunctionSpace<GFS> LFS;
  mutable LFS lfs;
  typedef Dune::PDELab::LFSIndexCache<LFS> LFSCache;
  mutable LFSCache lfsCache;
  typedef typename Container::template ConstLocalView<LFSCache> LView;
  mutable LView lView;

  mutable std::vector<RangeFieldType> lv;

  Entity entity;

  GFS_LocalViewer(GFS& gfs, Container& v) :
    lfs(gfs), lfsCache(lfs), lView(v), lv(gfs.maxLocalSize())
    {
    }

  template<typename EG, typename Femspace>
  bool DoSubEvaluation(const EG& eg, const Femspace& femspace)
  {
    if(entity == eg.entity())
    {
      return false;
    }
    entity = eg.entity();

    lfs.bind(entity);
    lfsCache.update();

    lView.bind(lfsCache);
    lView.read_sub_container(femspace,lv);
    lView.unbind();

    return true;
  }

  template<typename EG>
  bool DoEvaluation(const EG& eg)
  {
    if(entity == eg.entity())
    {
      return false;
    }
    entity = eg.entity();

    lfs.bind(entity);
    lfsCache.update();

    lView.bind(lfsCache);
    lView.read(lv);
    lView.unbind();

    return true;
  }

};

//_________________________________________________________________________________________________________________________

template<typename LFS>
class BasisFN
{
public:
  typedef typename LFS::Traits::FiniteElementType::Traits::LocalBasisType::Traits BTraits;
  const std::vector<typename BTraits::RangeType>& phi;
  const std::vector<typename BTraits::JacobianType>& gradphihat;
  std::vector<typename BTraits::JacobianType> gradphi;

  template<typename Pos, typename Cache, typename S>
  BasisFN(const Cache& cache, const Pos& position, const LFS& femspace, S& s) :
    phi(cache.evaluateFunction(position, femspace.finiteElement().localBasis())),
    gradphihat(cache.evaluateJacobian(position, femspace.finiteElement().localBasis())),
    gradphi(femspace.size())
    {
      for (size_t i=0; i<femspace.size(); i++)
      {
        s.mv(gradphihat[i][0],gradphi[i][0]);
      }
    }
};

//_________________________________________________________________________________________________________________________

template<typename RF, const int dim, typename CHLFS>
class CHVariables
{
public:

  const typename CHLFS::template Child<0>::Type& pfspace;
  const typename CHLFS::template Child<1>::Type& xispace;
  BasisFN<typename CHLFS::template Child<0>::Type> basis;

  RF pf;
  RF xi;
  Dune::FieldVector<RF,dim> gradpf;
  Dune::FieldVector<RF,dim> gradxi;

  template<typename Pos, typename Cache, typename LV, typename S>
  CHVariables(const Cache& cache, const Pos& position, const CHLFS& chLfs, const LV& lv, S& s) :
    pfspace(child(chLfs,Dune::Indices::_0)),
    xispace(child(chLfs,Dune::Indices::_1)),
    basis(cache, position, pfspace, s),
    pf(0.0),
    xi(0.0),
    gradpf(0.0),
    gradxi(0.0)
  {

    for (size_t i=0; i<pfspace.size(); i++)
    {
      pf += lv[pfspace.localIndex(i)]*basis.phi[i];
      xi += lv[xispace.localIndex(i)]*basis.phi[i];
      gradpf.axpy(lv[pfspace.localIndex(i)],basis.gradphi[i][0]);
      gradxi.axpy(lv[xispace.localIndex(i)],basis.gradphi[i][0]);
    }
  }
};


//_________________________________________________________________________________________________________________________

template<typename RF, const int dim, typename CHLFS>
class CHVariables_PFOld_Extension : public CHVariables<RF,dim,CHLFS>
{
public:


  RF pfOld;
  Dune::FieldVector<RF,dim> gradpfOld;


  template<typename Pos, typename Cache, typename LV, typename S>
  CHVariables_PFOld_Extension(const Cache& cache,
                            const Pos& position, const CHLFS& chLfs, const LV& lv, const LV& lvOld, S& s) :
    CHVariables<RF,dim,CHLFS>(cache, position, chLfs, lv, s),
    pfOld(0.0),
    gradpfOld(0.0)
  {
    for (size_t i=0; i<this->pfspace.size(); i++)
    {
      pfOld += lvOld[this->pfspace.localIndex(i)]*this->basis.phi[i];
      gradpfOld.axpy(lvOld[this->pfspace.localIndex(i)],this->basis.gradphi[i][0]);
    }
  }
};

//_________________________________________________________________________________________________________________________

template<typename RF, const int dim, typename CHLFS>
class CHVariables_U_PFOld_Extension : public CHVariables<RF,dim,CHLFS>
{
public:
  const typename CHLFS::template Child<2>::Type& uspace;

  RF pfOld;
  Dune::FieldVector<RF,dim> gradpfOld;
  RF u;
  Dune::FieldVector<RF,dim> gradu;
  RF uOld;


  template<typename Pos, typename Cache, typename LV, typename S>
  CHVariables_U_PFOld_Extension(const Cache& cache,
                            const Pos& position, const CHLFS& chLfs, const LV& lv, const LV& lvOld, S& s) :
    CHVariables<RF,dim,CHLFS>(cache, position, chLfs, lv, s),
    uspace(child(chLfs,Dune::Indices::_2)),
    pfOld(0.0),
    gradpfOld(0.0),
    u(0.0),
    gradu(0.0),
    uOld(0.0)
  {
    for (size_t i=0; i<this->pfspace.size(); i++)
    {
      u += lv[uspace.localIndex(i)]*this->basis.phi[i];
      gradu.axpy(lv[uspace.localIndex(i)],this->basis.gradphi[i][0]);
      uOld += lvOld [uspace.localIndex(i)]*this->basis.phi[i];
      pfOld += lvOld[this->pfspace.localIndex(i)]*this->basis.phi[i];
      gradpfOld.axpy(lvOld[this->pfspace.localIndex(i)],this->basis.gradphi[i][0]);
    }
  }
};


//_________________________________________________________________________________________________________________________

template<typename RF, const int dim, typename CHLFS>
class CHVariables_Threephase_Extension : public CHVariables<RF,dim,CHLFS>
{
public:
  const typename CHLFS::template Child<2>::Type& pf2space;
  const typename CHLFS::template Child<3>::Type& xi2space;

  RF pf2;
  Dune::FieldVector<RF,dim> gradpf2;
  RF xi2;
  Dune::FieldVector<RF,dim> gradxi2;

  template<typename Pos, typename Cache, typename LV, typename S>
  CHVariables_Threephase_Extension(const Cache& cache,
                            const Pos& position, const CHLFS& chLfs, const LV& lv, S& s) :
    CHVariables<RF,dim,CHLFS>(cache, position, chLfs, lv, s),
    pf2space(child(chLfs,Dune::Indices::_2)),
    xi2space(child(chLfs,Dune::Indices::_3)),
    pf2(0.0),
    gradpf2(0.0),
    xi2(0.0),
    gradxi2(0.0)
  {
    for (size_t i=0; i<this->pfspace.size(); i++)
    {
      pf2 += lv[pf2space.localIndex(i)]*this->basis.phi[i];
      gradpf2.axpy(lv[pf2space.localIndex(i)],this->basis.gradphi[i][0]);
      xi2 += lv[xi2space.localIndex(i)]*this->basis.phi[i];
      gradxi2.axpy(lv[xi2space.localIndex(i)],this->basis.gradphi[i][0]);
      }
  }
};


//_________________________________________________________________________________________________________________________

template<typename RF, const int dim, typename CHLFS>
class CHVariables_Threephase_PFOld_Extension : public CHVariables_Threephase_Extension<RF,dim,CHLFS>
{
public:

  RF pfOld;
  Dune::FieldVector<RF,dim> gradpfOld;
  RF pf2Old;
  Dune::FieldVector<RF,dim> gradpf2Old;


  template<typename Pos, typename Cache, typename LV, typename S>
  CHVariables_Threephase_PFOld_Extension(const Cache& cache,
                            const Pos& position, const CHLFS& chLfs, const LV& lv, const LV& lvOld, S& s) :
    CHVariables_Threephase_Extension<RF,dim,CHLFS>(cache, position, chLfs, lv, s),
    pfOld(0.0),
    gradpfOld(0.0),
    pf2Old(0.0),
    gradpf2Old(0.0)
  {
    for (size_t i=0; i<this->pfspace.size(); i++)
    {
      pfOld += lvOld[this->pfspace.localIndex(i)]*this->basis.phi[i];
      gradpfOld.axpy(lvOld[this->pfspace.localIndex(i)],this->basis.gradphi[i][0]);
      pf2Old += lvOld[this->pf2space.localIndex(i)]*this->basis.phi[i];
      gradpf2Old.axpy(lvOld[this->pf2space.localIndex(i)],this->basis.gradphi[i][0]);
    }
  }
};


template<typename RF, const int dim, typename CHLFS>
class CHVariables_Threephase_U_PFOld_Extension : public CHVariables_Threephase_Extension<RF,dim,CHLFS>
{
public:
  const typename CHLFS::template Child<4>::Type& uspace;

  RF pfOld;
  Dune::FieldVector<RF,dim> gradpfOld;
  RF pf2Old;
  Dune::FieldVector<RF,dim> gradpf2Old;
  RF u;
  Dune::FieldVector<RF,dim> gradu;
  RF uOld;


  template<typename Pos, typename Cache, typename LV, typename S>
  CHVariables_Threephase_U_PFOld_Extension(const Cache& cache,
                            const Pos& position, const CHLFS& chLfs, const LV& lv, const LV& lvOld, S& s) :
    CHVariables_Threephase_Extension<RF,dim,CHLFS>(cache, position, chLfs, lv, s),
    uspace(child(chLfs,Dune::Indices::_4)),
    pfOld(0.0),
    gradpfOld(0.0),
    pf2Old(0.0),
    gradpf2Old(0.0),
    u(0.0),
    gradu(0.0),
    uOld(0.0)
  {
    for (size_t i=0; i<this->pfspace.size(); i++)
    {
      u += lv[uspace.localIndex(i)]*this->basis.phi[i];
      gradu.axpy(lv[uspace.localIndex(i)],this->basis.gradphi[i][0]);
      uOld += lvOld [uspace.localIndex(i)]*this->basis.phi[i];
      pfOld += lvOld[this->pfspace.localIndex(i)]*this->basis.phi[i];
      gradpfOld.axpy(lvOld[this->pfspace.localIndex(i)],this->basis.gradphi[i][0]);
      pf2Old += lvOld[this->pf2space.localIndex(i)]*this->basis.phi[i];
      gradpf2Old.axpy(lvOld[this->pf2space.localIndex(i)],this->basis.gradphi[i][0]);
    }
  }
};


//_________________________________________________________________________________________________________________________

template<typename RF, const int dim, typename NSLFS>
class NSVariables
{
public:
  const typename NSLFS::template Child<0>::Type& vspace;
  const typename NSLFS::template Child<0>::Type::template Child<0>::Type& vfemspace;
  BasisFN<typename NSLFS::template Child<0>::Type::template Child<0>::Type> vbasis;

  Dune::FieldVector<RF,dim> v;
  Dune::FieldMatrix<RF,dim,dim> jacv;
  RF div_v;

  template<typename Pos, typename vCache, typename LV, typename S>
  NSVariables(const vCache& vcache, const Pos& position, const NSLFS& nsLfs, const LV& lv, S& s) :
    vspace(child(nsLfs,Dune::Indices::_0)),
    vfemspace(child(vspace,Dune::Indices::_0)),
    vbasis(vcache, position, vfemspace, s),
    v(0.0),
    jacv(0.0),
    div_v(0.0)
  {
    for(size_t k=0; k<dim; k++)
    {
      for (size_t i=0; i<vfemspace.size(); i++)
      {
        v[k] += lv[child(vspace,k).localIndex(i)]*vbasis.phi[i];
        jacv[k].axpy(lv[child(vspace,k).localIndex(i)],vbasis.gradphi[i][0]);
      }
      div_v += jacv[k][k];  // div is trace of jacobian
    }
  }
};
//_________________________________________________________________________________________________________________________

template<typename RF, const int dim, typename NSLFS>
class NSVariables_P_Extension : public NSVariables<RF,dim,NSLFS>
{
public:
  const typename NSLFS::template Child<1>::Type& pspace;
  BasisFN<typename NSLFS::template Child<1>::Type> pbasis;

  RF p;
  Dune::FieldVector<RF,dim> gradp;

  template<typename Pos, typename vCache, typename pCache, typename LV, typename S>
  NSVariables_P_Extension(const vCache& vcache, const pCache& pcache,
                            const Pos& position, const NSLFS& nsLfs, const LV& lv, S& s) :
    NSVariables<RF,dim,NSLFS>(vcache, position, nsLfs, lv, s),
    pspace(child(nsLfs,Dune::Indices::_1)),
    pbasis(pcache, position, pspace, s),
    p(0.0),
    gradp(0.0)
  {
    for (size_t i=0; i<pspace.size(); i++)
    {
      p += lv[pspace.localIndex(i)]*pbasis.phi[i];
      gradp.axpy(lv[pspace.localIndex(i)],pbasis.gradphi[i][0]);
    }
  }
};

//_________________________________________________________________________________________________________________________

template<typename RF, const int dim, typename NSLFS>
class NSVariables_P_VOld_Extension : public NSVariables_P_Extension<RF,dim,NSLFS>
{
public:
  Dune::FieldVector<RF,dim> vOld;

  template<typename Pos, typename vCache, typename pCache, typename LV, typename S>
  NSVariables_P_VOld_Extension(const vCache& vcache, const pCache& pcache,
                            const Pos& position, const NSLFS& nsLfs, const LV& lv, const LV& lvOld, S& s) :
    NSVariables_P_Extension<RF,dim,NSLFS>(vcache, pcache, position, nsLfs, lv, s),
    vOld(0.0)
  {

    for (size_t i=0; i<this->vfemspace.size(); i++)
    {
      for(size_t k=0; k<dim; k++)
      {
        vOld[k] += lvOld[child(this->vspace,k).localIndex(i)]*this->vbasis.phi[i];
      }
    }
  }
};
//_________________________________________________________________________________________________________________________


template<typename CHGFS>
using CHLocalBasisCache =
  Dune::PDELab::LocalBasisCache<typename CHGFS::template Child<0>::Type::Traits::FiniteElementType::Traits::LocalBasisType>;


template<typename NSGFS>
using VLocalBasisCache =
  Dune::PDELab::LocalBasisCache<typename NSGFS::template Child<0>::Type::template Child<0>::Type::Traits::FiniteElementType::Traits::LocalBasisType>;

template<typename NSGFS>
using PLocalBasisCache =
  Dune::PDELab::LocalBasisCache<typename NSGFS::template Child<1>::Type::Traits::FiniteElementType::Traits::LocalBasisType>;

// template<typename Param, typename CHGFS, typename CHContainer, typename NSGFS, typename NSContainer>
//     class FEM_CHNS_Base :
//       public Dune::PDELab::FullVolumePattern,
//       public Dune::PDELab::LocalOperatorDefaultFlags,
//       public Dune::PDELab::NumericalJacobianVolume<FEM_CHNS_Base<Param, CHGFS, CHContainer, NSGFS, NSContainer> >,
//       public Dune::PDELab::NumericalJacobianApplyVolume<FEM_CHNS_Base<Param, CHGFS, CHContainer, NSGFS, NSContainer> >
//     {
//     private:
//       typedef typename CHContainer::field_type RF;
//
//       typedef typename CHGFS::template Child<0>::Type FEM;
//       typedef typename FEM::Traits::FiniteElementType::Traits::LocalBasisType LocalBasis;
//       Dune::PDELab::LocalBasisCache<LocalBasis> cache;
//
//       typedef typename NSGFS::template Child<0>::Type::template Child<0>::Type VFEM;
//       typedef typename VFEM::Traits::FiniteElementType::Traits::LocalBasisType VLocalBasis;
//       Dune::PDELab::LocalBasisCache<VLocalBasis> vcache;
//
//       typedef typename NSGFS::template Child<1>::Type PFEM;
//       typedef typename PFEM::Traits::FiniteElementType::Traits::LocalBasisType PLocalBasis;
//       Dune::PDELab::LocalBasisCache<PLocalBasis> pcache;
//
//       Param& param; // parameter functions
//
//       //FOR EVALUATING the new and old cahn-Hilliard values
//       typedef Dune::PDELab::LocalFunctionSpace<CHGFS> CHLFS;
//       mutable CHLFS chLfs;
//       typedef Dune::PDELab::LFSIndexCache<CHLFS> CHLFSCache;
//       mutable CHLFSCache chLfsCache;
//       typedef typename CHContainer::template ConstLocalView<CHLFSCache> CHView;
//       mutable CHView chView;
//       mutable CHView chOldView;
//       mutable std::vector<typename LocalBasis::Traits::RangeFieldType> chLocal;
//       mutable std::vector<typename LocalBasis::Traits::RangeFieldType> chOldLocal;
//
//       //FOR EVALUATING the old navier-Stokes values
//       typedef Dune::PDELab::LocalFunctionSpace<NSGFS> NSLFS;
//       mutable NSLFS nsLfs;
//       typedef Dune::PDELab::LFSIndexCache<NSLFS> NSLFSCache;
//       mutable NSLFSCache nsLfsCache;
//       typedef typename NSContainer::template ConstLocalView<NSLFSCache> NSView;
//       mutable NSView nsView;
//       mutable std::vector<typename VLocalBasis::Traits::RangeFieldType> nsOldLocal;
//
//       //FOR DETECTING ENTITY changes
//       typedef typename NSGFS::Traits::GridViewType::template Codim<0>::Entity   ENT;
//       mutable ENT OldEnt;
//
//     public:
//       //! Boundary condition indicator type
//     //  using BC = StokesBoundaryCondition; // ??
//
//
// //      static const bool navier = P::assemble_navier;
// //      static const bool full_tensor = P::assemble_full_tensor;
//
//       // pattern assembly flags
//       enum { doPatternVolume = true };
//
//       // residual assembly flags
//       enum { doAlphaVolume = true };
//       //enum { doLambdaVolume = true };
//
//       //TODO enum { doLambdaBoundary = true };
//
//
//       FEM_CHNS_Base (Param& param_, CHGFS& chGfs_, CHContainer& chV_, CHContainer& chVOld_,
//                          NSGFS& nsGfs_, NSContainer& nsVOld_) :
//         param(param_),
//         chLfs(chGfs_), chLfsCache(chLfs), chView(chV_), chOldView(chVOld_),
//         chLocal(chGfs_.maxLocalSize()), chOldLocal(chGfs_.maxLocalSize()),
//         nsLfs(nsGfs_), nsLfsCache(nsLfs), nsView(nsVOld_), nsOldLocal(nsGfs_.maxLocalSize())
//       {
//       }
//
//
//       template<typename EG, typename LFSU, typename X, typename LFSV, typename R>
//       void alpha_volume (const EG& eg, const LFSU& lfsu, const X& x, const LFSV& lfsv, R& r) const
//       {
//         //Left hand side, terms depending on the solution x
//         using namespace Dune::Indices;
//
//         //Quadrature Rule
//         const int dim = EG::Entity::dimension;
//         const int order = 4; //TODO: Choose more educated
//         auto rule = Dune::PDELab::quadratureRule(eg.geometry(), order);
//         //const auto& myordering = lfsv.gridFunctionSpace().ordering();
//
//         const auto& pfspace = child(chLfs,_0);
//         const auto& xispace = child(chLfs,_1);
//         const auto& femspace = child(chLfs,_0);
//
//         const auto& vspace = child(lfsv,_0);
//         const auto& pspace = child(lfsv,_1);
//         const auto& vfemspace = child(vspace,_0);
//
//         if(eg.entity() != OldEnt)
//         {
//           OldEnt = eg.entity();
//
//           chLfs.bind(eg.entity());
//           chLfsCache.update();
//
//           chView.bind(chLfsCache);
//           chView.read(chLocal);
//           chView.unbind();
//
//           nsLfs.bind(eg.entity());
//           nsLfsCache.update();
//
//           nsView.bind(nsLfsCache);
//           nsView.read_sub_container(vspace,nsOldLocal);
//           nsView.unbind();
//         }
//
//         for(const auto& ip : rule)
//         {
//           // evaluate shape functions and  gradient of shape functions
//           auto& phi = cache.evaluateFunction(ip.position(), femspace.finiteElement().localBasis());
//           auto& gradphihat = cache.evaluateJacobian(ip.position(), femspace.finiteElement().localBasis());
//
//           auto& vphi = vcache.evaluateFunction(ip.position(), vfemspace.finiteElement().localBasis());
//           auto& gradvphihat = vcache.evaluateJacobian(ip.position(), vfemspace.finiteElement().localBasis());
//
//           auto& pphi = pcache.evaluateFunction(ip.position(), pspace.finiteElement().localBasis());
//           auto& gradpphihat = pcache.evaluateJacobian(ip.position(), pspace.finiteElement().localBasis());
//
//           // transform gradients of shape functions to real element
//           const auto S = eg.geometry().jacobianInverseTransposed(ip.position());
//           auto gradphi = makeJacobianContainer(femspace);
//           auto gradvphi = makeJacobianContainer(vfemspace);
//           auto gradpphi = makeJacobianContainer(pspace);
//           for (size_t i=0; i<femspace.size(); i++)
//           {
//             S.mv(gradphihat[i][0],gradphi[i][0]);
//           }
//           for (size_t i=0; i<vfemspace.size(); i++)
//           {
//             S.mv(gradvphihat[i][0],gradvphi[i][0]);
//           }
//           for (size_t i=0; i<pspace.size(); i++)
//           {
//             S.mv(gradpphihat[i][0],gradpphi[i][0]);
//           }
//
//           //compute current values of the unknowns
//           Dune::FieldVector<RF,dim> v(0.0);
//           Dune::FieldVector<RF,dim> vOld(0.0);
//           Dune::FieldMatrix<RF,dim,dim> jacv(0.0);
//           RF p=0.0;
//           Dune::FieldVector<RF,dim> gradp(0.0);
//           RF pf=0.0;
//           Dune::FieldVector<RF,dim> gradpf(0.0);
//           Dune::FieldVector<RF,dim> gradxi(0.0);
//
//           for (size_t i=0; i<pspace.size(); i++)
//           {
//             p += x(pspace,i)*pphi[i];
//             gradp.axpy(chLocal[pspace.localIndex(i)],gradpphi[i][0]);
//           }
//
//           for (size_t i=0; i<femspace.size(); i++)
//           {
//             pf += chLocal[pfspace.localIndex(i)]*phi[i];
//             gradpf.axpy(chLocal[pfspace.localIndex(i)],gradphi[i][0]);
//             gradxi.axpy(chLocal[xispace.localIndex(i)],gradphi[i][0]);
//           }
//           for (size_t i=0; i<vfemspace.size(); i++)
//           {
//             for(size_t k=0; k<dim; k++)
//             {
//               v[k] += x(child(vspace,k),i)*vphi[i];
//               vOld[k] += nsOldLocal[child(vspace,k).localIndex(i)]*vphi[i];
//               jacv[k].axpy(x(child(vspace,k),i),gradvphi[i][0]);
//             }
//           }
//
//           Dune::FieldVector<RF,dim> g(0.0);
//
//           //Dune::FieldVector<RF, dim> xg = eg.geometry().global( ip.position() );
//           //g[1]= 20*xg[0]*(1-xg[0]);
//           g[1] = -5;
//
//           RF rho_f=param.phys.rho_1*pf;
//           RF rho_f_reg= rho_f + param.delta;
//
//           RF pf_f = pf + 2*param.delta*(1-pf);
//           Dune::FieldVector<RF,dim> gradpf_f(0.0);
//           gradpf_f.axpy((1-2*param.delta),gradpf);
//
//           RF mu_reg = pf_f*param.phys.mu;
//
//           RF div_v = 0;
//           for(int k=0; k<dim; ++k)
//           {
//             div_v += jacv[k][k];    // divergence of v is the trace of the velocity jacobian
//           }
//
//           auto factor = ip.weight()*eg.geometry().integrationElement(ip.position());
//           for (size_t i=0; i<pspace.size(); i++)
//           {
//             //r.accumulate(pspace,i, -1.0* pf_f*(v*gradphi[i][0])*factor);
//             r.accumulate(pspace,i, -1.0 * (pf_f*div_v  + (gradpf_f*v) )* pphi[i] * factor);
//           }
//           for (size_t i=0; i<vfemspace.size(); i++)
//           {
//             for(int k=0; k<dim; k++)
//             {
//               //SIMPLIFY S=0, TODO: Implement this correctly
//
//               const auto v_nabla_v = v * jacv[k];
//
//
//               r.accumulate(child(vspace,k),i, (   rho_f_reg*(v[k]-vOld[k])*vphi[i]/param.time.dt          //TimeEvolution
//                                                 - p * (pf_f*gradvphi[i][0][k] + vphi[i]*gradpf_f[k])  //Pressure in weak form
//                                                 //+ pf_f * (gradp[k] * vphi[i])                           //Pressure in strong form
//                                                 + mu_reg * (jacv[k]*gradvphi[i][0])        //Viscosity
//                                                 + param.vDis.eval(pf_f)*v[k]*vphi[i]              //Velocity Dissipation in Solid
//                                                 //- rho*g[k]*vphi[i]                              //Gravitation
//                                                 //+ S                                             //Surface Tension
//                                                 //+ Reaction TODO
//                                               )*factor);
//             // integrate (grad u)^T * grad phi_i                                                  //Transpose for Viscosity
//             for(int dd=0; dd<dim; ++dd)
//                 r.accumulate(child(vspace,k),i, param.phys.mu * (jacv[dd][k] * (gradvphi[i][0][dd])) * factor);
//             }
//           }
//         }
//       }
//
//     };
