#pragma once

// C includes
#include<sys/stat.h>
// C++ includes
#include<iostream>
// Dune includes
#include<dune/grid/io/file/vtk.hh>
#include<dune/pdelab/newton/newton.hh>

#include<dune/pdelab/function/callableadapter.hh>
#include <dune/pdelab/common/functionutilities.hh>

#include <dune/phasefield/localoperator/fem_fs_chns_u_only.hh>
#include <dune/phasefield/localoperator/fem_fs_chns_driven_thinplateflow.hh>
#include <dune/phasefield/localoperator/fem_fs_r_only.hh>
#include <dune/phasefield/localoperator/pf_diff_estimator.hh>

#include <dune/phasefield/utility/ISTLBackend_NOVLP_BASE_PREC_FOR_GMRES.hh>

#include <dune/phasefield/chns_base.hh>
#include <dune/phasefield/fs_chns_r.hh>
#include <dune/phasefield/initial_fn.hh>
#include <dune/phasefield/boundary_fn.hh>
#include <dune/phasefield/debug/vector_debug.hh>

#include "summerschool_timestepmanager.hh"
#include "boundary_summerschool.hh"



//_________________________________________________________________________________________________________________________

template<typename Grid, typename GV,typename ES, typename V_FEM, typename P_FEM, typename PHI_FEM>
void driver_fs (Grid& grid, GV& gv, const ES& es, const V_FEM& vFem, const P_FEM pFem, const PHI_FEM& phiFem, Dune::ParameterTree& ptree, Dune::MPIHelper& helper)
{


  const int dim = GV::dimension;
  typedef double RF;


  typedef Params_summerschool< RF,
                        DoubleWell_limited< RF, DoubleWell_poly<RF>, Limiter_FakeDiverging<RF> >,//DoubleWell_poly<RF>,
                        Mobility_Const<RF>,
                        Reaction_Calcite_Precipitation<RF>,
                        VelDissipation_Quadratic<RF> > Parameters;
  Parameters param(ptree);
  SummerschoolTimestepManager timestepManager(ptree.sub("Time"),param.time,2);


//_________________________________________________________________________________________________________________________
// Make grid function spaces


  typedef GFS_fs_chns_r<RF, ES, V_FEM, P_FEM, PHI_FEM, PHI_FEM> GFS;
  GFS gfs(es,vFem,pFem,phiFem,phiFem);
  typedef typename GFS::NS NS_GFS;
  typedef typename GFS::CH CH_GFS;


  //_________________________________________________________________________________________________________________________
  //Constraints

  typedef Bdry_fs_chns_r_compositeV_2d<v_per,
              Dune::PDELab::DirichletConstraintsParameters,
              p_per,
              InflowOutflow_phi_ss,
              Dune::PDELab::NoDirichletConstraintsParameters,
              InflowOutflow_u_ss> Bdry;
  Bdry bdry;
  gfs.updateConstraintsContainer(bdry);

  //_________________________________________________________________________________________________________________________
  //Coefficient vectors

  using CH_V = Dune::PDELab::Backend::Vector<CH_GFS, RF>;
  using NS_V = Dune::PDELab::Backend::Vector<NS_GFS, RF>;
  CH_V chV(gfs.ch);
  NS_V nsV(gfs.ns);

  //_________________________________________________________________________________________________________________________
  //Initial Values

    typedef Ini_fs_chns_r< VelocityInitial_Inflow_ss<GV,RF,Parameters,dim>,
                      PInitial_Inflow<GV,RF,Parameters>,
                      PhiInitial_SummerSchoolPaper<GV,RF,Parameters>,
                      ZeroInitial<GV,RF,Parameters>,
                      UInitial_Inflow<GV,RF,Parameters> >  Initial;
    Initial initial(gv,param);
    Dune::PDELab::interpolate(initial.iniCH , gfs.ch, chV);
    Dune::PDELab::interpolate(initial.iniNS , gfs.ns, nsV);

  //____________________________________________________________________________________________________________________
  // Refine Initial Data

  for(int i=0; i<ptree.get("Refinement.maxRefineLevel",2); i++)
  {
    // mark elements for refinement
    typedef RefinementIndicator<GV,RF,PF_DIFF_ESTIMATOR<CH_GFS>,CH_GFS> RIND0;
    RIND0 rInd0(gv,gfs.ch,ptree.sub("Refinement"));
    rInd0.calculateIndicator(gv,chV,false);
    rInd0.markGrid(grid,2);
    //rInd0.refineGrid(grid,gv,chV,nsV,gfs.ch,gfs.ns,1);
    rInd0.myRefineGrid(grid,gv,chV,nsV,gfs.ch,gfs.ns,1);
    Dune::PDELab::interpolate(initial.iniCH , gfs.ch, chV);
    Dune::PDELab::interpolate(initial.iniNS , gfs.ns, nsV);

    //vtkwriter.write(param.time.t + i,Dune::VTK::ascii);
  }

  gfs.updateConstraintsContainer(bdry);
  interpolateToBdry(gfs,initial.iniNS,initial.iniCH,nsV,chV);

  CH_V chVOld = chV;
  NS_V nsVOld = nsV;




  //_________________________________________________________________________________________________________________________
  // LocalOperators

  typedef FEM_FS_CHNS_Driven_ThinPlateFlow<Parameters, CH_GFS, CH_V, NS_GFS, NS_V> NS_LOP;
  NS_LOP nsLop(param, gfs.ch, chV, gfs.ns, nsVOld);

  typedef FEM_FS_CHNS_U_Only<Parameters, CH_GFS, CH_V, NS_GFS, NS_V> CH_LOP;
  CH_LOP chLop(param, gfs.ch, chVOld, gfs.ns, nsV);


  typedef Dune::PDELab::ISTL::BCRSMatrixBackend<> MBE;
  MBE mbe(10);// Maximal number of nonzeroes per row can be cross-checked by patternStatistics().


  //_________________________________________________________________________________________________________________________
  // Construct discrete Grid Functions
  typedef DGF_fs_chns_r<GFS, NS_V, CH_V> DGF;
  DGF dgf(gfs, nsV, chV);

  //_________________________________________________________________________________________________________________________
  // prepare VTK writer
  std::string filename=ptree.get("output.filename","output");
  struct stat st;
  if( stat( filename.c_str(), &st ) != 0 )
  {
    int stat = 0;
    stat = mkdir(filename.c_str(),S_IRWXU|S_IRWXG|S_IRWXO);
    if( stat != 0 && stat != -1)
      std::cout << "Error: Cannot create directory "  << filename << std::endl;
  }
  auto stationaryvtkwriter = std::make_shared<Dune::SubsamplingVTKWriter<GV> >(gv,1);
  Dune::VTKSequenceWriter<GV> vtkwriter(stationaryvtkwriter,filename.c_str(),filename.c_str(),"");
  dgf.addToVTKWriter(vtkwriter);



  vtkwriter.write(param.time.t,Dune::VTK::base64);

  //_________________________________________________________________________________________________________________________
  printVector(chV,10,"chV");
  printVector(nsV,10,"nsV");


  //____________________________________________________________________________________________________________________
  //int i_it=0;

  bool FirstReactionStep = true;

  while (!timestepManager.isFinished())
    {

      // Grid Operators
      typedef Dune::PDELab::GridOperator<NS_GFS,NS_GFS,NS_LOP,MBE,RF,RF,RF,typename GFS::NS_CC,typename GFS::NS_CC> NS_GO;
      NS_GO nsGo(gfs.ns,gfs.nsCC,gfs.ns,gfs.nsCC,nsLop,mbe);

      typedef Dune::PDELab::GridOperator<CH_GFS,CH_GFS,CH_LOP,MBE,RF,RF,RF,typename GFS::CH_CC,typename GFS::CH_CC> CH_GO;
      CH_GO chGo(gfs.ch,gfs.chCC,gfs.ch,gfs.chCC,chLop,mbe);

      // Linear solver
      typedef ISTLBackend_NOVLP_BASE_PREC_FOR_GMRES
                  <CH_GO, Dune::SeqILU0, Dune::RestartedGMResSolver> CH_LS; CH_LS ls_ch(chGo,1000,200,3,1);

      typedef ISTLBackend_NOVLP_BASE_PREC_FOR_GMRES
                  <NS_GO, Dune::SeqILU0, Dune::RestartedGMResSolver> NS_LS; NS_LS ls_ns(nsGo,4000,400,3,1);


      // Nonlinear solver
      typedef Dune::PDELab::Newton<NS_GO,NS_LS,NS_V> PDESolver1;
      PDESolver1 ns_newton(nsGo,nsV,ls_ns);
      ns_newton.setParameters(ptree.sub("NS_Newton"));

      typedef Dune::PDELab::Newton<CH_GO,CH_LS,CH_V> PDESolver2;
      PDESolver2 ch_newton(chGo,chV,ls_ch);
      ch_newton.setParameters(ptree.sub("Phasefield_Newton"));

      //_________________________________________________________________________________________________________________________
      // do time step


      interpolateToBdry(gfs,initial.iniNS,initial.iniCH, nsV, chV);

      int result;
      if((param.time.t > ptree.get("Time.flowUntilStable",(double)5) )
        )
      {
        if(FirstReactionStep)
        {
          auto velocityFunction = Dune::PDELab::makeGridFunctionFromCallable (gv, [&](const auto& element, const auto& xlocal){
              Dune::FieldVector<double,1> y1;
              dgf.phiDgf.evaluate(element, xlocal, y1);
              Dune::FieldVector<double,dim> y2;
              dgf.vDgf.evaluate(element, xlocal, y2);
              Dune::FieldVector<double,1> result;
              result[0] = y1[0]*y2[0];
              return result;
          });
          Dune::FieldVector<double,1> integral;
          Dune::PDELab::integrateGridFunction(velocityFunction,integral,2);
          std::cout << std::endl<< std::endl <<"Velocity Integral: " << integral[0]<< std::endl<< std::endl<< std::endl;

          nsV *= 0.353/integral[0];
          nsVOld *= 0.353/integral[0];

          FirstReactionStep = false;
        }

        result = timestepManager.doOnlyCHTimestep(ch_newton,vtkwriter);

        auto cfluxFunction = Dune::PDELab::makeGridFunctionFromCallable (gv, [&](const auto& element, const auto& xlocal){
          Dune::FieldVector<double,1> y0;
          dgf.uDgf.evaluate(element, xlocal, y0);
          Dune::FieldVector<double,1> y1;
          dgf.phiDgf.evaluate(element, xlocal, y1);
          Dune::FieldVector<double,dim> y2;
          dgf.vDgf.evaluate(element, xlocal, y2);
          Dune::FieldVector<double,1> result;
          result[0] = y0[0]*y1[0]*y2[0];
          return result;
        });
        Dune::FieldVector<double,1> cfintegral;
        Dune::PDELab::integrateGridFunction(cfluxFunction,cfintegral,2);
        std::cout << "CFlux Integral: " << cfintegral[0]<< "  CFlux/Velocity :" << cfintegral[0]/0.353 << std::endl;
      }
      else
      {
        result = timestepManager.doOnlyNSTimestep(ns_newton,vtkwriter);
      }

      if(result <0)
      {
        return;
      }
      else if(result == 0)
      {
        nsV = nsVOld;
        chV = chVOld;
        continue;
      }

      chVOld = chV;
      nsVOld = nsV;

      //_________________________________________________________________________________________________________________________

      // recompute constraints
      gfs.updateConstraintsContainer(bdry);
      interpolateToBdry(gfs,initial.iniNS,initial.iniCH,nsV,chV);
      interpolateToBdry(gfs,initial.iniNS,initial.iniCH,nsVOld,chVOld);


      //vtkwriter.write(timestepManager.t+2e-6,Dune::VTK::ascii);
    }
}
